﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WMM.Tapioca.Linkbase;
using System.Xml.Linq;
using WMM.Tapioca.Formula.Filters;
using WMM.Tapioca.Instance;
using Saxon.Api;
using WMM.Tapioca.Utilities;

namespace WMM.Tapioca.Formula
{
    /// <summary>
    /// An existence assertion is a statement about the number of times that the assertion's 
    /// variable set will evaluate against a input XBRL instance.
    /// </summary>
    public class ExistenceAssertion : VariableSetAssertion
    {
        /// <summary>
        /// Constructor.
        /// </summary>
        public ExistenceAssertion(Resource resourceAssertion, Dictionary<string, string> namespaces)
            : base(resourceAssertion, namespaces)
        {
            foreach (XAttribute attr in resourceAssertion.ResourceXElement.Attributes())
            {
                if (attr.Name.LocalName.Equals("aspectModel"))
                {
                    AspectModel = attr.Value;
                }

                if (attr.Name.LocalName.Equals("implicitFiltering"))
                {
                    ImplicitFiltering = Boolean.Parse(attr.Value);
                }

                if (attr.Name.LocalName.Equals("test"))
                {
                    Test = attr.Value;
                }

                if (attr.Name.LocalName.Equals("id"))
                {
                    Id = attr.Value;
                }
            }

            foreach (Arc arc in resourceAssertion.ExtendedLinkParent.Arcs)
            {
                if (arc.FromId.Equals(Label) &&
                    arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2008/element-label"))
                {
                    Linkbase.Resource resource = GetResource(arc.ToId, resourceAssertion.ExtendedLinkParent);
                    if (resource.LocalName.Equals("label"))
                    {
                        ElementLabel = resource.Content;
                    }
                }

                if (arc.FromId.Equals(Label) &&
                    arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2010/assertion-satisfied-message"))
                {
                    AssertionToMessageArcs.Add(arc);
                    Linkbase.Resource resource = GetResource(arc.ToId, resourceAssertion.ExtendedLinkParent);
                    if (resource.LocalName.Equals("message"))
                    {
                        MessageSatisfied = new Message(resource);
                    }
                }

                if (arc.FromId.Equals(Label) &&
                    arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2010/assertion-unsatisfied-message"))
                {
                    AssertionToMessageArcs.Add(arc);

                    Linkbase.Resource resource = GetResource(arc.ToId, resourceAssertion.ExtendedLinkParent);

                    if (resource.LocalName.Equals("message"))
                    {
                        MessageUnsatisfied = new Message(resource);
                    }
                }

                if (arc.FromId.Equals(Label) && arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2008/variable-set"))
                {
                    VariableSetArcs.Add(new VariableSet(arc));

                    List<Resource> resources = GetResources(arc.ToId, resourceAssertion.ExtendedLinkParent);

                    foreach (Resource resource in resources)
                    {

                        if (resource.LocalName.Equals("parameter"))
                        {
                            Parameters.Add(new Parameter(resource));
                        }

                        if (resource.LocalName.Equals("factVariable"))
                        {
                            FactVariables.Add(new FactVariable(resource));
                        }

                        if (resource.LocalName.Equals("generalVariable"))
                        {
                            GeneralVariables.Add(new GeneralVariable(resource));
                        }
                    }
                }

                if (arc.FromId.Equals(Label) && arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2008/variable-set-precondition"))
                {
                    Linkbase.Resource resource = GetResource(arc.ToId, resourceAssertion.ExtendedLinkParent);

                    if (resource.LocalName.Equals("precondition"))
                    {
                        Precondition = new Precondition(resource);
                    }
                }

                if (arc.FromId.Equals(Label) && arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2008/element-label"))
                {
                    LabelAssertion = GetResource(arc.ToId, resourceAssertion.ExtendedLinkParent);
                }
            }

            foreach (Arc arc in resourceAssertion.ExtendedLinkParent.Arcs)
            {
                if (arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2008/variable-filter"))
                {
                    foreach (FactVariable factVariable in FactVariables)
                    {
                        if (factVariable.Label.Equals(arc.FromId))
                        {
                            List<Resource> resources = GetResources(arc.ToId, arc.ExtendedLinkParent);

                            foreach (Resource resource in resources)
                            {
                                Filter filter = Filter.CreateFilter(arc, resource, this);
                                factVariable.Filters.Add(filter);
                                factVariable.VariableFilterArcs.Add(arc);

                                if (filter.Cover)
                                {
                                    AddCoveredAspect(factVariable, arc);
                                }
                            }
                        }
                    }
                }
            }

            foreach (Arc arc in resourceAssertion.ExtendedLinkParent.Arcs)
            {
                if (arc.ArcRole.AbsoluteUri.Equals("http://xbrl.org/arcrole/2008/variable-set-filter"))
                {                   
                    Resource filterResource = GetResource(arc.ToId, arc.ExtendedLinkParent);
                    GroupFilters.Add(Filter.CreateFilter(arc, filterResource, this));                     
                }
            }

            if (test == null)
            {
                test = ". gt 0";
            }

        }


        /// <summary>
        /// Evaluating the assertion.
        /// </summary>
        public override bool EvaluateAssertion(InstanceDocument instance, FormulaProcessor formulaProcessor)
        {
            int k = 1;

            this.dts = instance.Dts;
            namespaces = dts.DtsFormula.Namespaces;

            // clear
            variableItems.Clear();
            variableAsSequenceEntries.Clear();
            variableAsItem.Clear();
            variableNamesAsItem.Clear();
            parameterDictionary.Clear();
            dependencyGraph = new Graph();

            // Declare parameter
            DeclareParameters(this, compiler);

            foreach (VariableSet variableSet in this.VariableSetArcs)
            {
                if (variableSet.Variable.Name.LocalName.Equals("factVariable") ||
                    variableSet.Variable.Name.LocalName.Equals("generalVariable"))
                {
                    if (test.Contains(variableSet.Name))
                    {
                        log.Error("Error: The Existence Assertion: " + this.Label + " . the existence-assertion expression has a variable dependency on a general variable or a fact variable.");
                        dts.ValidatedFormula = false;
                        return false;
                    }
                }
            }

            XPathSelector selector = compiler.Compile(test).Load();            

            BindParameters(selector, this, parameterDictionary);

            Dictionary<string, XdmAtomicValue> variableFallBack = new Dictionary<string, XdmAtomicValue>();

            // Evaluate variable set

            try
            {
                EvaluateFactVariables(this.variableItems, this.variableSequences, variableFallBack);
            }
            catch (Exception ex)
            {
                log4net.GlobalContext.Properties["assertionId"] = this.Id + " (evaluation " + k + ")";
                log4net.GlobalContext.Properties["satisfied"] = "Exception";
                FormulaProcessor.log.Error("Assertion:" + this.Id + " message:" + ex.Message);
                formulaProcessor.assertionsNotSatisfied++;
                k = k + 1;
                return false;
            }

            for (int i = 0; i < this.variableItems.Count; i++)
            {
                variableAsItem.Add(this.variableItems.ElementAt(i).Value);
                variableNamesAsItem.Add(this.variableItems.ElementAt(i).Key);
            }

            List<List<FactItem>> cartesianProductItems = ConvertToList(MyExtensions.CartesianProduct(variableAsItem).ToList());

            Dictionary<string, FactItem> variableNotSequenceEntries = new Dictionary<string, FactItem>();

            variableAsSequenceEntries.Clear();

            if (this.ImplicitFiltering && this.variableItems.Count > 1)
            {
                foreach (List<FactItem> cartesianProductItem in cartesianProductItems)
                {
                    if (this.ImplicitlyMatchItems(cartesianProductItem, variableNamesAsItem, this.variableSequences, this))
                    {
                        selector.ContextItem = new XdmAtomicValue(cartesianProductItems.Count);
                    }
                    else
                    {
                        selector.ContextItem = new XdmAtomicValue(0);
                    }

                    log4net.GlobalContext.Properties["assertionId"] = this.Id + " (evaluation " + k + ")";

                    bool evaluateResult = Boolean.Parse(selector.Evaluate().ToString());

                    BuildMessage(formulaProcessor, compiler, parameterDictionary, variableNotSequenceEntries, evaluateResult, variableFallBack);

                    k = k + 1;
                }
            }
            else
            {
                selector.ContextItem = new XdmAtomicValue(cartesianProductItems.Count);

                log4net.GlobalContext.Properties["assertionId"] = this.Id + " (evaluation " + k + ")";

                bool evaluateResult = Boolean.Parse(selector.Evaluate().ToString());

                BuildMessage(formulaProcessor, compiler, parameterDictionary, variableNotSequenceEntries, evaluateResult, variableFallBack);

                k = k + 1;
            }

            bool result = false;

            if (cartesianProductItems.Count == 0)
            {
                Dictionary<string, FactItem> variableEntries2 = new Dictionary<string, FactItem>();

                selector.ContextItem = new XdmAtomicValue(0);

                log4net.GlobalContext.Properties["assertionId"] = this.Id + " (evaluation " + k + ")";

                result = Boolean.Parse(selector.Evaluate().ToString());

                BuildMessage(formulaProcessor, compiler, parameterDictionary, variableEntries2, result, variableFallBack);                
            }

            return result;
        }


    }
}
