﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using WMM.Tapioca.Utilities;

namespace WMM.Tapioca.Linkbase
{
    /// <summary>
    /// XBRL 2.1 Specification 3.5.2.5
    /// The arcroleRef element is used to resolve custom xlink:arcrole values that
    /// are used in a linkbase or an XBRL instance (for footnoteArc). 
    /// </summary>
    [Serializable]
    public class ArcRoleRef
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        internal string href;
        internal string schemaUri;
        internal string id;
        internal string arcroleURI;
        internal XbrlDocument parentDocument;

        /// <summary>
        /// Creates a new instance of <see cref="ArcRoleRef"/>.
        /// </summary>
        public ArcRoleRef()
        {
        }

        /// <summary>
        /// Overloaded.  Creates a new instance of <see cref="ArcRoleRef"/>.
        /// </summary>
        /// <param name="document">The parent document.</param>
        /// <param name="hrefArg">The href property value to be assigned 
        /// to the newly created <see cref="ArcRoleRef"/>.</param>
        /// <param name="uriArg">The uri property value to be assigned 
        /// to the newly created <see cref="ArcRoleRef"/>.</param>
        public ArcRoleRef(XbrlDocument document, string hrefArg, string uriArg)
        {
            log.Debug("Creating ArcRoleRef: " + uriArg + "-" + hrefArg);

            href = hrefArg;
            schemaUri = GetSchemaName();
            id = GetId();
            arcroleURI = uriArg;
            parentDocument = document;
        }


        /// <summary>
        /// Retrieves the schema portion of this <see cref="ArcRoleRef"/>'s href property.
        /// </summary>
        /// <example>
        /// If href = "balancesheet.xsd#assets", GetSchemaName = "balancesheet.xsd".
        /// </example>
        /// <returns>The parsed schema.</returns>
        /// <exception cref="ApplicationException">If the href is not well formed.</exception>
        /// <exception cref="ArgumentNullException">If the href is null.</exception>
        internal string GetSchemaName()
        {
            string[] strs = href.Split('#');

            if (strs.Length < 2)
            {
                throw new XbrlException("Can't split href: " + href);
            }

            return strs[0];
        }

        /// <summary>
        /// Retrieves the id portion of this <see cref="ArcRoleRef"/>'s href property.
        /// </summary>
        /// <example>
        /// If href = "balancesheet.xsd#assets", GetSchemaName = "assets".
        /// </example>
        /// <exception cref="ApplicationException">If the href is not well formed.</exception>
        /// <exception cref="ArgumentNullException">If the href is null.</exception>
        public string GetId()
        {
            string[] strs = href.Split('#');

            if (strs.Length < 2)
            {
                throw new XbrlException("Can't split href: " + href);
            }

            return strs[1];
        }

        internal string GetHref()
        {
            return href;
        }

        internal void SetHref(string s)
        {
            href = s;
        }

    }
}
