﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.Xml;
using System.Xml.Schema;
using System.Xml.Linq;

namespace WMM.Tapioca.Taxonomy
{
    /// <summary>
    /// XBRL 2.1 Specification 5.1.4
    /// An encapsulation of the XBRL element "ArcroleType". 
    /// </summary>
    public class ArcRoleType
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        /// <summary>
        /// Get the cyclesAllowed attribute value.
        /// </summary>
        public string CyclesAllowed
        {
            get;
            private set;
        }

        /// <summary>
        /// The schema that references this role type.
        /// </summary>
        public XmlSchema Schema
        {
            get;
            private set;
        }

        /// <summary>
        /// The URI for this role.
        /// </summary>
        public Uri ArcRoleUri
        {
            get;
            private set;
        }

        /// <summary>
        /// The ID of this role type.
        /// </summary>
        public string Id
        {
            get;
            private set;
        }

        /// <summary>
        /// The definition of this role type.
        /// </summary>
        public string Definition
        {
            get;
            private set;
        }

        /// <summary>
        /// A collection of "UsedOn" references for this role type. Used to identify what
        /// elements may use a taxonomy defined role or arc role value.
        /// </summary>
        public List<XName> UsedOnReferences
        {
            get;
            private set;
        }


        internal ArcRoleType(XmlSchema containingXbrlSchema, XmlNode arcRoleTypeNode)
        {
            this.Schema = containingXbrlSchema;
            this.UsedOnReferences = new List<XName>();

            try
            {
                this.ArcRoleUri = new Uri(XmlUtilities.GetAttributeValue(arcRoleTypeNode, "arcroleURI"));
            }
            catch (Exception ex)
            {
                log.Error("Validation error on schema:", ex);
            }

            this.Id = XmlUtilities.GetAttributeValue(arcRoleTypeNode, "id");
            this.CyclesAllowed = XmlUtilities.GetAttributeValue(arcRoleTypeNode, "cyclesAllowed");

            log.Debug("Creating ArcRoleType: " + this.ArcRoleUri + "-" + this.Id + " CyclesAllowed:" + this.CyclesAllowed);

            foreach (XmlNode currentNode in arcRoleTypeNode.ChildNodes)
            {
                if (currentNode.LocalName.Equals("definition"))
                {
                    this.Definition = currentNode.InnerText;
                }

                if (currentNode.LocalName.Equals("usedOn"))
                {
                    string valueQN = currentNode.InnerText;

                    string prefix = GetPrefix(valueQN);
                    string name = "";

                    foreach (XmlQualifiedName nameSpace in containingXbrlSchema.Namespaces.ToArray())
                    {
                        if (nameSpace.Name.Equals(prefix))
                        {
                            name = nameSpace.Namespace;
                            break;
                        }
                    }

                    if (name.Equals(""))
                    {
                        name = currentNode.GetNamespaceOfPrefix(prefix);
                    }

                    XNamespace ns = name;
                    XName value = ns + GetLocalName(valueQN);
                    this.UsedOnReferences.Add(value);
                }
            }
        }


        private string GetLocalName(string valueQN)
        {
            string[] dimAttrQN = valueQN.Split(':');

            if (dimAttrQN.Count() > 1)
            {
                return dimAttrQN[1];
            }
            else
            {
                return dimAttrQN[0];
            }

        }


        private string GetPrefix(string valueQN)
        {
            string[] dimAttrQN = valueQN.Split(':');

            if (dimAttrQN.Count() > 1)
            {
                return dimAttrQN[0];
            }
            else
            {
                return "";
            }

        }


    }
}
