using System;
using System.Text;
using System.Xml;
using System.Xml.Schema;
using WMM.Tapioca.Taxonomy;
using WMM.Tapioca.Utilities;

namespace WMM.Tapioca.Taxonomy
{
	/// <summary>
    /// XBRL 2.1 Specification 5.1.1
    /// Each concept defined in a taxonomy schema is uniquely identified by an elements syntax definition in the taxonomy schema. 
    /// To correspond to a concept definition, an XML Schema element definition has to specify the elements name, a substitution group, 
    /// and type. All element names MUST be unique within a given taxonomy schema. 
	/// </summary>
    public class Concept : IIndexable
    {
        private static readonly log4net.ILog log = log4net.LogManager.GetLogger(System.Reflection.MethodBase.GetCurrentMethod().DeclaringType);

        int IIndexable.Id
        {
            get { return GetHashCode(); }
        }

        int IIndexable.Name
        {
            get { return GetNameHashCode(schemaElement.QualifiedName.Namespace, schemaElement.QualifiedName.Name); }
        }

		/// <summary>
		/// The set of possible substitution groups for an element.
		/// </summary>
        public enum EnumSubstitutionGroup
        {
			/// <summary>
			/// An unknown substitution group.
			/// </summary>
            Unknown,
			/// <summary>
			/// An item substitution group.
			/// </summary>
            Item,
			/// <summary>
			/// A tuple substitution group.
			/// </summary>
            Tuple,
			/// <summary>
			/// A dimension item substitution group.
			/// </summary>
            DimensionItem,
			/// <summary>
			/// A hypercube item substitution group.
			/// </summary>
            HypercubeItem
        }

		/// <summary>
		/// The set of possible period types for an element.
		/// </summary>
        public enum EnumPeriodType
        {
			/// <summary>
			/// An unknown period type.
			/// </summary>
            Unknown,
			/// <summary>
			/// An instant period type.
			/// </summary>
            Instant,
			/// <summary>
			/// A duration period type.
			/// </summary>
            Duration
        }

        /// <summary>
        /// The set of possible balance for an element.
        /// </summary>
        public enum EnumBalanceType
        {
            /// <summary>
            /// An unknown period type.
            /// </summary>
            Unknown,
            /// <summary>
            /// A credit type.
            /// </summary>
            Credit,
            /// <summary>
            /// A debit type.
            /// </summary>
            Debit
        }        

		/// <summary>
		/// The name of the element.
		/// </summary>
		public string Name
		{
			get;
			private set;
		}

		/// <summary>
		/// The ID of the element.
		/// </summary>
		public string Id
		{
			get;
			private set;
		}

		/// <summary>
		/// The substitution group for the element.
		/// </summary>
		public EnumSubstitutionGroup SubstitutionGroup
		{
			get;
			private set;
		}

        private XmlSchemaElement schemaElement;

        /// <summary>
        /// The element from schema.
        /// </summary>
        public XmlSchemaElement SchemaElement
        {
            get
            {
                return schemaElement;
            }
        }

        private string periodTypeString;

        /// <summary>
        /// The period type for the element.
        /// </summary>
        public string PeriodTypeString
        {
            get { return periodTypeString; }
            set { periodTypeString = value; }
        }

		/// <summary>
		/// The period type for the element.
		/// </summary>
		public EnumPeriodType PeriodType
		{
			get;
			private set;
		}

        private string balanceString;

        /// <summary>
        /// The balance for the element.
        /// </summary>
        public string BalanceString
        {
            get { return balanceString; }
            set { balanceString = value; }
        }

        /// <summary>
        /// The balance type for the element.
        /// </summary>
        public EnumBalanceType BalanceType
        {
            get;
            private set;
        }

		/// <summary>
		/// A reference to the schema which contains the element.
		/// </summary>
		public TaxonomySchema Schema
		{
			get;
			private set;
		}

		/// <summary>
		/// The type name of this element.
		/// </summary>
		public XmlQualifiedName TypeName
		{
			get;
			private set;
		}

        /// <summary>
        /// The type of this item.
        /// </summary>
        public XmlSchemaType Type
        {
            get;
            private set;
        }



        internal Concept(TaxonomySchema schema, XmlSchemaElement schemaElement)
        {
            this.Schema = schema;
            this.schemaElement = schemaElement;
            this.Id = schemaElement.Id;
            this.Name = schemaElement.Name;
            this.TypeName = schemaElement.SchemaTypeName;
            this.Type = schemaElement.ElementSchemaType;

            SetPeriodType();
            SetBalance();
            SetSubstitutionGroup(schemaElement.SubstitutionGroup);
        }


		/// <summary>
		/// Determines whether or not a supplied object equals the current element.
		/// </summary>
		/// <param name="obj">
		/// The object to compare to the current element.
		/// </param>
		/// <returns>
		/// Returns true if the supplied object equals the current element. Returns false if
		/// the supplied object does not equal the current element.
		/// </returns>
        public override bool Equals(object obj)
        {
            if (!(obj is Concept))
            {
                return false;
            }

            Concept otherElement = obj as Concept;

            return otherElement.Id.Equals(this.Id);
        }		


        private void SetPeriodType()
        {
            this.PeriodType = EnumPeriodType.Unknown;

            if (schemaElement.UnhandledAttributes != null)
            {
                foreach (XmlAttribute currentAttribute in schemaElement.UnhandledAttributes)
                {
                    if (currentAttribute.LocalName.Equals("periodType") &&
                        currentAttribute.NamespaceURI.Equals(Constants.XBRLI))
                    {
                        periodTypeString = currentAttribute.Value;
                        SetPeriodType(currentAttribute.Value);
                        return;
                    }
                }
            }
        }


        private void SetPeriodType(string periodType)
        {
            if (periodType.Equals("instant"))
            {
                this.PeriodType = EnumPeriodType.Instant;
            }
            else if (periodType.Equals("duration"))
            {
                this.PeriodType = EnumPeriodType.Duration;
            }
            else
            {
                // We can't identify the type, so throw an exception.

                string messageFormat = AssemblyResources.GetName("InvalidElementPeriodType");
                StringBuilder messageBuilder = new StringBuilder();
                messageBuilder.AppendFormat(messageFormat, this.Schema.PathFile, periodType, this.Name);
                log.Error("Validation error on concept", new XbrlException(messageBuilder.ToString()));
                Schema.Dts.ValidatedInstance = false;
            }
        }


        private void SetBalance()
        {           
            this.BalanceType = EnumBalanceType.Unknown;
            this.balanceString = "";

            if (schemaElement.UnhandledAttributes != null)
            {
                foreach (XmlAttribute currentAttribute in schemaElement.UnhandledAttributes)
                {
                    if (currentAttribute.LocalName.Equals("balance") &&
                        currentAttribute.NamespaceURI.Equals(Constants.XBRLI))
                    {
                        balanceString = currentAttribute.Value;
                        SetBalanceType(currentAttribute.Value);
                        return;
                    }
                }
            }
        }


        private void SetBalanceType(string balanceType)
        {
            if (balanceType.Equals("credit"))
            {
                this.BalanceType = EnumBalanceType.Credit;
            }
            else if (balanceType.Equals("debit"))
            {
                this.BalanceType = EnumBalanceType.Debit;
            }
            else
            {
                log.Error("Validation error on concept", new XbrlException("Invalid Element Balance Type"));
                Schema.Dts.ValidatedInstance = false;
            }
        }

 
        private void SetSubstitutionGroup(XmlQualifiedName substitutionGroupValue)
        {
            this.SubstitutionGroup = EnumSubstitutionGroup.Unknown;

            if ((substitutionGroupValue.Name.Length == 0) && (substitutionGroupValue.Namespace.Length == 0))
            {
                return;
            }

            if ((substitutionGroupValue.Namespace.Equals(Constants.XBRLI)) && (substitutionGroupValue.Name.Equals("item")))
            {
                this.SubstitutionGroup = EnumSubstitutionGroup.Item;
            }
            else if ((substitutionGroupValue.Namespace.Equals(Constants.XBRLI)) && (substitutionGroupValue.Name.Equals("tuple")))
            {
                this.SubstitutionGroup = EnumSubstitutionGroup.Tuple;
            }
            else if ((substitutionGroupValue.Namespace.Equals(Constants.XBRLDT)) && (substitutionGroupValue.Name.Equals("dimensionItem")))
            {
                this.SubstitutionGroup = EnumSubstitutionGroup.DimensionItem;
            }
            else if ((substitutionGroupValue.Namespace.Equals(Constants.XBRLDT)) && (substitutionGroupValue.Name.Equals("hypercubeItem")))
            {
                this.SubstitutionGroup = EnumSubstitutionGroup.HypercubeItem;
            }
            else
            {
                // We can't identify the type, so throw an exception.

                
            }
        }

        /// <summary>
        /// Calculates a hash code for the element.
        /// </summary>
        /// <returns>
        /// A hash code for the element.
        /// </returns>
        //public override int GetHashCode()
        //{
        //    return this.Id.GetHashCode();
        //}


        public override int GetHashCode()
        {
            if (Id != null)
            {
                return this.Id.GetHashCode();
            }

            int hash = 1;
            hash = hash * 31 + (schemaElement.QualifiedName.Name != null ? schemaElement.QualifiedName.Name.GetHashCode() : 0);
            hash = hash * 31 + (schemaElement.QualifiedName.Namespace != null ? schemaElement.QualifiedName.Namespace.GetHashCode() : 0);
            return hash;
        }

        public int GetNameHashCode(string ns, string name)
        {
            int hash = 1;
            hash = hash * 51 + (name != null ? name.GetHashCode() : 0);
            hash = hash * 51 + (ns != null ? ns.GetHashCode() : 0);
            return hash;
        }
    }
}
