using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;

namespace WMM.Test.Tapioca
{
	/// <summary>
	/// A variation of a test case as defined in an XBRL-CONF-CR3-2007-03-05 test case document.
	/// </summary>
	/// <remarks>
	/// <para>
	/// Test case variations are found in one of the conformance suite test documents.
	/// A test case document is an XML document with the following form:
	/// </para>
	/// <para>
	/// &lt;testcase xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" name="Identifier Scope" description="Section 4.3  The Item Concept" outpath="out" owner="mg@fla.fujitsu.com" xsi:noNamespaceSchemaLocation="../lib/test.xsd" minimal="true"&gt;
	/// &lt;variation id="..." name="..."&gt;
	/// &lt;description&gt;...&lt;/description&gt;
	/// &lt;data&gt;
	/// &lt;xsd readMeFirst="[true|false]"&gt;...&lt;/xsd&gt;
	/// &lt;instance readMeFirst="[true|false]"&gt;...&lt;/instance&gt;
	/// &lt;/data&gt;
	/// &lt;result expected="[valid|invalid]"/&gt;
	/// &lt;/variation&gt;
	/// ... other &lt;testcase&gt; nodes ...
	/// &lt;/testcase&gt;
	/// </para>
	/// <para>
	/// This class represents one of the &lt;variation&gt; nodes as well as the content in its child nodes.
	/// </para>
	/// </remarks>
    internal class TestVariation
    {
		internal string Id { get; private set; }
        internal string Name { get; private set; }
        internal string Description { get; private set; }
        internal string Xsd { get; private set; }
        internal string Linkbase { get; private set; }
        internal string Instance { get; private set; }
        internal string Error { get; private set; }
        internal bool isReadXsdFirst { get; private set; }
        internal bool isReadLinkbaseFirst { get; set; }
        internal bool isReadInstanceFirst { get; private set; }
        internal bool isValidityExpected { get; private set; }

        internal TestVariation(XmlNode VariationNode)
        {
            Error = "";
            XmlNamespaceManager nsmgr = new XmlNamespaceManager(VariationNode.OwnerDocument.NameTable);
            nsmgr.AddNamespace("c", "http://xbrl.org/2008/conformance");
            var descriptionNode = VariationNode.SelectSingleNode("child::description");
            var dataNode = VariationNode.SelectSingleNode("child::data");
            var resultNode = VariationNode.SelectSingleNode("child::result");
            var xsdNode = dataNode.SelectSingleNode("child::xsd");
            var linkbaseNode = dataNode.SelectSingleNode("child::linkbase");
            var instanceNode = dataNode.SelectSingleNode("child::instance");

            this.Id = VariationNode.Attributes["id"].Value;
            this.Name = VariationNode.Attributes["name"].Value;
            this.Description = descriptionNode.InnerText;

            if (resultNode.Attributes["expected"] != null)
            {
                string ExpectedResultText = resultNode.Attributes["expected"].Value;

                if (ExpectedResultText == "valid")
                {
                    this.isValidityExpected = true;
                }
                else
                {
                    this.isValidityExpected = false;
                }
            }

            if (resultNode.HasChildNodes)
            {
                XmlNode firstNode = resultNode.FirstChild;

                if (firstNode.LocalName.Equals("error"))
                {
                    Error = firstNode.InnerXml;
                }
            }

            if (xsdNode != null)
            {
                this.Xsd = xsdNode.InnerText;
                this.isReadXsdFirst = false;
                if (xsdNode.Attributes["readMeFirst"] != null)
                {
                    string ReadXsdFirstText = xsdNode.Attributes["readMeFirst"].Value;

                    if (ReadXsdFirstText == "true")
                    {
                        this.isReadXsdFirst = true;
                    }
                    else
                    {
                        this.isReadXsdFirst = false;
                    }
                }
            }
            else
            {
                this.Xsd = string.Empty;
                this.isReadXsdFirst = false;
            }

            if (instanceNode != null)
            {
                this.Instance = instanceNode.InnerText;
                string ReadInstanceFirstText = instanceNode.Attributes["readMeFirst"].Value;

                if (ReadInstanceFirstText == "true")
                {
                    this.isReadInstanceFirst = true;
                }
                else
                {
                    this.isReadInstanceFirst = false;
                }
            }
            else
            {
                 this.Instance = string.Empty;
            }

            if (linkbaseNode != null)
            {
                this.Linkbase = linkbaseNode.InnerText;
                this.isReadLinkbaseFirst = false;
                if (linkbaseNode.Attributes["readMeFirst"] != null)
                {
                    string ReadLinkbaseFirstText = linkbaseNode.Attributes["readMeFirst"].Value;

                    if (ReadLinkbaseFirstText == "true")
                    {
                        this.isReadLinkbaseFirst = true;
                    }
                    else
                    {
                        this.isReadLinkbaseFirst = false;
                    }
                }
            }            
        }

       
    }
}
